"""Type aliases used by the public API"""

import sys
from dataclasses import dataclass
from typing import Callable, List, Mapping, MutableMapping, Optional, TypeVar, Union

if sys.version_info < (3, 8):
    from typing_extensions import Literal  # this supports python < 3.8
else:
    from typing import Literal  # this is available in python 3.8+


@dataclass
class ValidationSuccess:
    valid: Literal[True] = True


@dataclass
class ValidationFailure:
    error: str
    valid: Literal[False] = False


Answer = Union[bool, str]
Answers = Mapping[str, Answer]
MutableAnswers = MutableMapping[str, Answer]
OptionList = List[str]
V = TypeVar("V")
StaticOrDynamicValue = Union[V, Callable[[Answers], V]]
ShouldAsk = Callable[[Answers], bool]
ValidationResponse = Union[ValidationSuccess, ValidationFailure]
Validator = Callable[[str, Answers], Union[ValidationResponse, Optional[str]]]



from argparse import Namespace

import pytest

from columbo import (
    BasicQuestion,
    Choice,
    CliException,
    Confirm,
    DuplicateQuestionNameException,
    Echo,
    parse_args,
)
from columbo._cli import create_parser, format_cli_help, to_answers
from columbo._interaction import canonical_arg_name
from tests.sample_data import (
    DUPLICATE_QUESTION_NAME_PARAMS,
    SOME_ARG_NAME,
    SOME_BOOL,
    SOME_DEFAULT,
    SOME_INVALID_ARG_NAME,
    SOME_INVALID_OPTION,
    SOME_NAME,
    SOME_NAMESPACE,
    SOME_NON_DEFAULT_OPTION,
    SOME_OPTIONS,
    SOME_OTHER_BOOL,
    SOME_OTHER_STRING,
    SOME_STRING,
    SampleQuestion,
    some_dynamic_options,
)

invalid_cli_parameterize = pytest.mark.parametrize(
    ("description", "questions", "args"),
    [
        (
            "invalid static arg, correct arg name",
            [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)],
            [SOME_ARG_NAME, SOME_INVALID_OPTION],
        ),
        (
            "invalid dynamic arg, correct arg name",
            [Choice(SOME_NAME, SOME_STRING, some_dynamic_options, SOME_DEFAULT)],
            [SOME_ARG_NAME, SOME_INVALID_OPTION],
        ),
        (
            "invalid static arg, incorrect arg name",
            [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)],
            [SOME_INVALID_ARG_NAME, SOME_NON_DEFAULT_OPTION],
        ),
        (
            "invalid dynamic arg, incorrect arg name",
            [Choice(SOME_NAME, SOME_STRING, some_dynamic_options, SOME_DEFAULT)],
            [SOME_INVALID_ARG_NAME, SOME_NON_DEFAULT_OPTION],
        ),
    ],
)


@invalid_cli_parameterize
def test_parse_args__invalid_arg_default_exit_on_error__system_exit(
    questions, args, description
):
    with pytest.raises(SystemExit):
        parse_args(questions, args)


@invalid_cli_parameterize
def test_parse_args__invalid_arg_exit_on_error__system_exit(
    questions, args, description
):
    with pytest.raises(SystemExit):
        parse_args(questions, args)


@invalid_cli_parameterize
def test_parse_args__invalid_arg_no_exit_on_error__system_exit(
    questions, args, description
):
    with pytest.raises(CliException):
        parse_args(questions, args, exit_on_error=False)


def test_parse_args__valid_arg__answers():
    questions = [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)]

    answers = parse_args(questions, [SOME_ARG_NAME, SOME_NON_DEFAULT_OPTION])

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_NON_DEFAULT_OPTION


def test_parse_args__initial_answers__answers():
    initial_key = "initial-key"
    initial_answers = {initial_key: "aa-test-bb"}
    questions = [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)]

    answers = parse_args(
        questions, [SOME_ARG_NAME, SOME_NON_DEFAULT_OPTION], answers=initial_answers
    )

    assert answers[initial_key] == initial_answers[initial_key]


@pytest.mark.parametrize("questions", DUPLICATE_QUESTION_NAME_PARAMS)
def test_parse_args__duplicate_question_name__exception(questions):
    with pytest.raises(DuplicateQuestionNameException):
        parse_args(questions)


def test_parse_args__duplicate_question_name_in_answers__exception():
    with pytest.raises(DuplicateQuestionNameException):
        parse_args(
            [BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)],
            answers={SOME_NAME: "existing value"},
        )


def test_parse_args__parser_name__name_in_result(capsys):
    try:
        parse_args(
            [Confirm(SOME_NAME, SOME_STRING)],
            [SOME_INVALID_ARG_NAME, SOME_INVALID_OPTION],
            parser_name=SOME_OTHER_STRING,
        )
    except SystemExit:
        assert SOME_OTHER_STRING in capsys.readouterr().err
    else:
        pytest.fail("System exit should have been thrown")


def test_create_parser__choice_not_given__no_value():
    parser = create_parser([Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)])

    result = parser.parse_args([])

    assert vars(result)[SOME_NAME] is None


@pytest.mark.parametrize("choice", SOME_OPTIONS)
def test_create_parser__choice__choice_valid(choice):
    parser = create_parser([Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)])

    result = parser.parse_args([SOME_ARG_NAME, choice])

    assert vars(result)[SOME_NAME] == choice


def test_create_parser__choice_invalid_option__exception():
    parser = create_parser([Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)])

    with pytest.raises(SystemExit):
        parser.parse_args([SOME_ARG_NAME, SOME_INVALID_OPTION])


def test_create_parser__choice_dynamic_options_invalid_option__value_stored():
    # dynamic options can't be statically validated, so we store any value
    parser = create_parser(
        [Choice(SOME_NAME, SOME_STRING, some_dynamic_options, SOME_DEFAULT)]
    )

    result = parser.parse_args([SOME_ARG_NAME, SOME_INVALID_OPTION])

    assert vars(result)[SOME_NAME] == SOME_INVALID_OPTION


def test_create_parser__confirm_not_given__no_value():
    parser = create_parser([Confirm(SOME_NAME, SOME_STRING)])

    result = parser.parse_args([])

    assert vars(result)[SOME_NAME] is None


def test_create_parser__confirm_main_option__true():
    parser = create_parser([Confirm(SOME_NAME, SOME_STRING)])

    result = parser.parse_args([SOME_ARG_NAME])

    assert vars(result)[SOME_NAME] is True


def test_create_parser__confirm_no_option__false():
    parser = create_parser([Confirm(SOME_NAME, SOME_STRING)])

    result = parser.parse_args([canonical_arg_name(f"no-{SOME_NAME}")])

    assert vars(result)[SOME_NAME] is False


def test_create_parser__basic_question_not_given__no_value():
    parser = create_parser([BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)])

    result = parser.parse_args([])

    assert vars(result)[SOME_NAME] is None


def test_create_parser__basic_question__value():
    parser = create_parser([BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)])

    result = parser.parse_args([SOME_ARG_NAME, SOME_STRING])

    assert vars(result)[SOME_NAME] is SOME_STRING


def test_create_parser__unsupported_question__exception():
    with pytest.raises(ValueError):
        create_parser([SampleQuestion(SOME_NAME, SOME_STRING)])


def test_to_answer__basic_question_value_not_set__default():
    questions = [BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)]

    answers = to_answers(questions, Namespace())

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_DEFAULT


def test_to_answer__basic_question_value_set__set_value():
    questions = [BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)]

    answers = to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_STRING


def test_to_answer__basic_question_value_not_valid__exception():
    questions = [
        BasicQuestion(
            SOME_NAME, SOME_STRING, SOME_DEFAULT, validator=lambda v, a: "some-error"
        )
    ]

    with pytest.raises(CliException):
        with pytest.deprecated_call():
            to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))


def test_to_answer__basic_question_dont_ask__value_not_stored():
    questions = [
        BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT, should_ask=lambda _: False)
    ]

    result = to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))

    assert SOME_NAME not in result


def test_to_answer__choice_value_not_set__default():
    questions = [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)]

    answers = to_answers(questions, Namespace())

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_DEFAULT


def test_to_answer__choice_value_set__set_value():
    questions = [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)]

    answers = to_answers(questions, Namespace(**{SOME_NAME: SOME_NON_DEFAULT_OPTION}))

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_NON_DEFAULT_OPTION


def test_to_answer__choice_value_not_valid__exception():
    questions = [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)]

    with pytest.raises(CliException):
        to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))


def test_to_answer__choice_dont_ask__value_not_stored():
    questions = [
        Choice(
            SOME_NAME,
            SOME_STRING,
            SOME_OPTIONS,
            SOME_DEFAULT,
            should_ask=lambda _: False,
        )
    ]

    result = to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))

    assert SOME_NAME not in result


def test_to_answer__confirm_value_not_set__default():
    questions = [Confirm(SOME_NAME, SOME_STRING, default=SOME_BOOL)]

    answers = to_answers(questions, Namespace())

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_BOOL


def test_to_answer__confirm_value_set__set_value():
    questions = [Confirm(SOME_NAME, SOME_STRING, default=SOME_BOOL)]

    answers = to_answers(questions, Namespace(**{SOME_NAME: SOME_OTHER_BOOL}))

    assert SOME_NAME in answers
    assert answers[SOME_NAME] == SOME_OTHER_BOOL


def test_to_answer__confirm_dont_ask__value_not_stored():
    questions = [Confirm(SOME_NAME, SOME_STRING, should_ask=lambda _: False)]

    result = to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))

    assert SOME_NAME not in result


def test_to_answer__echo__value_not_stored():
    questions = [Echo(SOME_STRING)]

    result = to_answers(questions, Namespace(**{SOME_NAME: SOME_STRING}))

    assert SOME_NAME not in result


def test_to_answer__unsupported_question__exception():
    with pytest.raises(ValueError):
        to_answers([SampleQuestion(SOME_NAME, SOME_STRING)], SOME_NAMESPACE)


@pytest.mark.parametrize(
    ["description", "questions", "partial_expected_result"],
    [
        (
            "basic question",
            [BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)],
            "--my-test-value MY-TEST-VALUE",
        ),
        (
            "choice static options, options listed",
            [Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)],
            "--my-test-value {x,y,z}",
        ),
        (
            "choice dynamic options, options not listed",
            [Choice(SOME_NAME, SOME_STRING, some_dynamic_options, SOME_DEFAULT)],
            "--my-test-value MY-TEST-VALUE",
        ),
        (
            "confirm, enabled option",
            [Confirm(SOME_NAME, SOME_STRING)],
            "--my-test-value",
        ),
        (
            "confirm, disabled option",
            [Confirm(SOME_NAME, SOME_STRING)],
            "--no-my-test-value",
        ),
    ],
)
def test_format_cli_help__questions__contains_expected_result(
    questions, partial_expected_result, description
):
    result = format_cli_help(questions)

    assert partial_expected_result in result, description


def test_format_cli_help__parser_name__name_in_result():
    result = format_cli_help(
        [Confirm(SOME_NAME, SOME_STRING)], parser_name=SOME_OTHER_STRING
    )

    assert SOME_OTHER_STRING in result


@pytest.mark.parametrize("questions", DUPLICATE_QUESTION_NAME_PARAMS)
def test_format_cli_help__duplicate_question_name__exception(questions):
    with pytest.raises(DuplicateQuestionNameException):
        format_cli_help(questions)



import pytest

from columbo._exception import (
    CliException,
    ColumboException,
    DuplicateQuestionNameException,
)


@pytest.mark.parametrize(
    "exception",
    [CliException("test-value", "--test-argument"), DuplicateQuestionNameException],
)
def test_all_custom_exception__inherits_from_custom_base(exception):
    with pytest.raises(ColumboException):
        raise exception



import re
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from typing import Collection, Optional, Type, TypeVar, Union

from columbo import _user_io as user_io
from columbo._exception import DuplicateQuestionNameException
from columbo._types import (
    Answer,
    Answers,
    MutableAnswers,
    OptionList,
    ShouldAsk,
    StaticOrDynamicValue,
    V,
    ValidationFailure,
    ValidationResponse,
    ValidationSuccess,
    Validator,
)

Interaction = Union["Echo", "Acknowledge", "Question"]


# Used by copy() implementations. Since some arguments can be None, None can't be used as the value to indicate that the
# the argument was not given.
class _Sentinel(Enum):
    A = 0


T = TypeVar("T")
_NOT_GIVEN = _Sentinel.A
Possible = Union[T, _Sentinel]

VALIDATOR_UPGRADE_DOCS_LINK = "https://wayfair-incubator.github.io/columbo/0.10.0/usage-guide/validators/#upgrading-validator-structure"


# value is Possible[T]. The type is explicitly not annotated because of a conflict when T is a union. The type system
# flattens unions of unions. This causes the type system to infer that T is object instead of the nested union type.
def _or_default(value, default: T) -> T:
    return default if isinstance(value, _Sentinel) else value


class Echo:
    """Display a message to the user."""

    def __init__(self, message: StaticOrDynamicValue[str]) -> None:
        """
        Initialize an instance.

        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        """
        self._message = message

    def display(self, answers: Answers) -> None:
        user_io.echo(to_value(self._message, answers, str))

    def copy(
        self, *, message: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN
    ) -> "Echo":
        """
        Create a new instance like this one, potentially with different values.

        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :return: A newly constructed instance with the given values in place of the values of this instance.
        """
        return Echo(_or_default(message, self._message))


class Acknowledge:
    """Display a message to the user and require the user to press ENTER to continue."""

    def __init__(self, message: StaticOrDynamicValue[str]) -> None:
        """
        Initialize an instance.

        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        """
        self._message = message

    def display(self, answers: Answers) -> None:
        user_io.acknowledge(to_value(self._message, answers, str))

    def copy(
        self, *, message: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN
    ) -> "Acknowledge":
        """
        Create a new instance like this one, potentially with different values.

        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :return: A newly constructed instance with the given values in place of the values of this instance.
        """
        return Acknowledge(_or_default(message, self._message))


class Question(ABC):
    """
    Base class for a prompt to the user that produces an answer.
    """

    def __init__(
        self,
        name: str,
        message: StaticOrDynamicValue[str],
        cli_help: Optional[str] = None,
        should_ask: Optional[ShouldAsk] = None,
    ) -> None:
        """
        Initialize an instance.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        """
        self._name = name
        self._message = message
        self._cli_help = cli_help
        self._should_ask = should_ask

    @property
    def name(self) -> str:
        return self._name

    @property
    def cli_help(self) -> Optional[str]:
        return self._cli_help

    @abstractmethod
    def ask(self, answers: Answers, no_user_input: bool = False) -> Answer:
        """
        Prompt the user with this question.

        :param answers: The answers that have been provided this far.
        :return: The answer to the question.
        :param no_user_input: If `True` the default value for the question will be used without waiting for the user
            to provide an answer. Default: False
        :return: The answer to the question.
        """
        pass

    def should_ask(self, answers: Answers) -> bool:
        """
        Should the user be asked this question.

        :param answers: The answers that have been provided this far.
        :return: `True` if this questions should be asked
        """
        if self._should_ask is None:
            return True
        if callable(self._should_ask):
            return self._should_ask(answers)
        raise ValueError(f"Invalid value for should_ask: {self._should_ask}")


class Confirm(Question):
    """
    A question with a yes or no answer.
    """

    def __init__(
        self,
        name: str,
        message: StaticOrDynamicValue[str],
        default: StaticOrDynamicValue[bool] = False,
        cli_help: Optional[str] = None,
        should_ask: Optional[ShouldAsk] = None,
    ) -> None:
        """
        Initialize an instance.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param default: The default answer to the question. If the value is callable, the argument passed in will be the
            answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        """
        super().__init__(name, message, cli_help, should_ask)
        self._default = default

    @property
    def default(self) -> StaticOrDynamicValue[bool]:
        return self._default

    def ask(self, answers: Answers, no_user_input: bool = False) -> bool:
        """
        Prompt the user with this question.

        :param answers: The answers that have been provided this far.
        :param no_user_input: If `True` the default value for the question will be used without waiting for the user
            to provide an answer. Default: False
        :return: The answer to the question.
        """
        return user_io.confirm(
            to_value(self._message, answers, str),
            default=to_value(self._default, answers, bool),
            no_user_input=no_user_input,
        )

    def copy(
        self,
        *,
        name: Possible[str] = _NOT_GIVEN,
        message: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN,
        default: Possible[StaticOrDynamicValue[bool]] = _NOT_GIVEN,
        cli_help: Possible[Optional[str]] = _NOT_GIVEN,
        should_ask: Possible[Optional[ShouldAsk]] = _NOT_GIVEN,
    ) -> "Confirm":
        """
        Create a new instance like this one, potentially with different values.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param default: The default answer to the question. If the value is callable, the argument passed in will be the
            answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        :return: A newly constructed instance with the given values in place of the values of this instance.
        """
        return Confirm(
            _or_default(name, self._name),
            _or_default(message, self._message),
            _or_default(default, self._default),
            _or_default(cli_help, self._cli_help),
            _or_default(should_ask, self._should_ask),
        )


class Choice(Question):
    """
    A question with a set of possible answers.
    """

    def __init__(
        self,
        name: str,
        message: StaticOrDynamicValue[str],
        options: StaticOrDynamicValue[OptionList],
        default: StaticOrDynamicValue[str],
        cli_help: Optional[str] = None,
        should_ask: Optional[ShouldAsk] = None,
    ) -> None:
        """
        Initialize an instance.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param options: The set of possible answers to the question. If the value is callable, the argument passed in
            will be the answers that have been provided this far.
        :param default: The default answer to the question. If the value is callable, the argument passed in will be the
            answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        """
        super().__init__(name, message, cli_help, should_ask)
        self._options = options
        self._default = default

    @property
    def options(self) -> StaticOrDynamicValue[OptionList]:
        return self._options

    @property
    def default(self) -> StaticOrDynamicValue[str]:
        return self._default

    def validate(self, value: str, answers: Answers) -> ValidationResponse:
        """Validate the value (a new answer).

        :param value: The identifier that will be used as the key to access the this question's answer.
        :param answers: The answers that have been provided this far.
        :return: A ValidationFailure or ValidationSuccess object.
        """
        options = to_value(self._options, answers, list)
        if value not in options:
            return ValidationFailure(error=f"Chosen value: {value} not in options")
        return ValidationSuccess()

    def ask(self, answers: Answers, no_user_input: bool = False) -> str:
        """
        Prompt the user with this question.

        :param answers: The answers that have been provided this far.
        :param no_user_input: If `True` the default value for the question will be used without waiting for the user
            to provide an answer. Default: False
        :return: The answer to the question.
        """
        return user_io.multiple_choice(
            to_value(self._message, answers, str),
            to_value(self._options, answers, list),
            default=to_value(self._default, answers, str),
            no_user_input=no_user_input,
        )

    def copy(
        self,
        *,
        name: Possible[str] = _NOT_GIVEN,
        message: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN,
        options: Possible[StaticOrDynamicValue[OptionList]] = _NOT_GIVEN,
        default: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN,
        cli_help: Possible[Optional[str]] = _NOT_GIVEN,
        should_ask: Possible[Optional[ShouldAsk]] = _NOT_GIVEN,
    ) -> "Choice":
        """
        Create a new instance like this one, potentially with different values.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param options: The set of possible answers to the question. If the value is callable, the argument passed in
            will be the answers that have been provided this far.
        :param default: The default answer to the question. If the value is callable, the argument passed in will be the
            answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        :return: A newly constructed instance with the given values in place of the values of this instance.
        """
        return Choice(
            _or_default(name, self._name),
            _or_default(message, self._message),
            _or_default(options, self._options),
            _or_default(default, self._default),
            _or_default(cli_help, self._cli_help),
            _or_default(should_ask, self._should_ask),
        )


class BasicQuestion(Question):
    """
    A question with an arbitrary text answer.
    """

    def __init__(
        self,
        name: str,
        message: StaticOrDynamicValue[str],
        default: StaticOrDynamicValue[str],
        cli_help: Optional[str] = None,
        should_ask: Optional[ShouldAsk] = None,
        validator: Optional[Validator] = None,
    ) -> None:
        """
        Initialize an instance.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param default: The default answer to the question. If the value is callable, the argument passed in will be the
            answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        :param validator: Callable that will validate the response given by the user.
            A ValidationSuccess object indicates success and a ValidationFailure object indicates failure.
        """
        super().__init__(name, message, cli_help, should_ask)
        self._default = default
        self._validator = validator

    @property
    def default(self) -> StaticOrDynamicValue[str]:
        return self._default

    def validate(self, value: str, answers: Answers) -> ValidationResponse:
        """Validate the value (a new answer).

        :param value: The identifier that will be used as the key to access the this question's answer.
        :param answers: The answers that have been provided this far.
        :return: A ValidationFailure or ValidationSuccess object.
        """
        if self._validator is None:
            return ValidationSuccess()
        if callable(self._validator):
            result = self._validator(value, answers)
            if isinstance(result, (ValidationFailure, ValidationSuccess)):
                return result
            else:  # handle deprecated, legacy validator responses
                warnings.warn(
                    f"You are using a validator {self._validator} that returns content in a way is deprecated and will be removed after the 1.0.0 release."
                    + f"For information on how to upgrade, please refer to the documentation here: {VALIDATOR_UPGRADE_DOCS_LINK}",
                    DeprecationWarning,
                )
                if result:
                    return ValidationFailure(error=result)
                return ValidationSuccess()
        raise ValueError(f"Invalid value for validate: {self._validator}")

    def ask(self, answers: Answers, no_user_input: bool = False) -> str:
        """
        Prompt the user with this question.

        :param answers: The answers that have been provided this far.
        :param no_user_input: If `True` the default value for the question will be used without waiting for the user
            to provide an answer. Default: False
        :return: The answer to the question.
        """

        # ask question until answer is valid
        while True:
            answer = user_io.ask(
                to_value(self._message, answers, str),
                default=to_value(self._default, answers, str),
                no_user_input=no_user_input,
            )

            result = self.validate(answer, answers)
            if result.valid:
                break

            user_io.echo(
                f"The answer you have provided is not valid:\n{result.error}\n\n"
                "We will continue asking questions until you provide a valid answer"
            )

        return answer

    def copy(
        self,
        *,
        name: Possible[str] = _NOT_GIVEN,
        message: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN,
        default: Possible[StaticOrDynamicValue[str]] = _NOT_GIVEN,
        cli_help: Possible[Optional[str]] = _NOT_GIVEN,
        should_ask: Possible[Optional[ShouldAsk]] = _NOT_GIVEN,
        validator: Possible[Optional[Validator]] = _NOT_GIVEN,
    ) -> "BasicQuestion":
        """
        Create a new instance like this one, potentially with different values.

        :param name: The identifier that will be used as the key to access the this question's answer.
        :param message: The message to be displayed to the user. If the value is callable, the argument passed in will
            be the answers that have been provided this far.
        :param default: The default answer to the question. If the value is callable, the argument passed in will be the
            answers that have been provided this far.
        :param cli_help: Optional help message to be displayed for command line interface.
        :param should_ask: If `None`, the question is asked. Otherwise, the callable will be passed the answers that
            have been provided this far and should return `True` if the question should be asked.
        :param validator: Callable that will validate the response given by the user.
            None indicates that validation was successful. Otherwise, a string containing details
            of the error that caused the validation failure.
        """
        return BasicQuestion(
            _or_default(name, self._name),
            _or_default(message, self._message),
            _or_default(default, self._default),
            _or_default(cli_help, self._cli_help),
            _or_default(should_ask, self._should_ask),
            _or_default(validator, self._validator),
        )


def to_value(
    value: StaticOrDynamicValue[V], answers: Answers, value_type: Type[V]
) -> V:
    if isinstance(value, value_type):
        return value
    if callable(value):
        return value(answers)
    raise ValueError(f"Invalid value: {value}")


def get_answers(
    interactions: Collection[Interaction],
    answers: Optional[Answers] = None,
    no_user_input: bool = False,
) -> MutableAnswers:
    """
    Iterates over collection of interactions, invoking interaction specific behavior.

    :param interactions: Collection of interactions to present the user with.
    :param answers: An initial dictionary of answers to start from.
    :param no_user_input: If `True` the default value for the question will be used without waiting for the user
        to provide an answer. Default: False
    :return: Dictionary of answers.
    :raises DuplicateQuestionNameException: One of the given questions attempts to reuse a name. This includes a
        question that was used to create `answers` if given.
    """
    validate_duplicate_question_names(interactions, answers)
    result = {} if answers is None else dict(answers)

    for interaction in interactions:
        if isinstance(interaction, (Echo, Acknowledge)):
            interaction.display(result)
        elif isinstance(interaction, Question):
            if interaction.should_ask(result):
                result[interaction.name] = interaction.ask(result, no_user_input)
        else:
            raise ValueError(f"Unsupported interaction type: {type(interaction)}")

    return result


def canonical_arg_name(name: str) -> str:
    sanizized_name = name.lower().replace(" ", "-").replace("_", "-").strip("-")
    # remove any duplicate dashes ("foo--bar" becomes "foo-bar")
    arg_name = re.sub(r"(\-)\1+", r"\1", sanizized_name)
    return f"--{arg_name}"


def validate_duplicate_question_names(
    interactions: Collection[Interaction], answers: Optional[Answers] = None
) -> None:
    """
    Ensure that multiple questions don't use the same name.

    :param interactions: Collection of interactions to check.
    :param answers: An initial dictionary of answers to treat as questions that have already been asked.
    :raises DuplicateQuestionNameException: One of the given questions attempts to reuse a name.
    """
    used_names = set() if answers is None else set(answers.keys())
    for interaction in interactions:
        if isinstance(interaction, Question):
            name_variants = {interaction.name, canonical_arg_name(interaction.name)}
            if name_variants.intersection(used_names):
                raise DuplicateQuestionNameException(
                    f"{interaction.name} has already been used"
                )
            used_names.update(name_variants)



"""
Helpful wrappers for prompt-toolkit functionality.
"""

from typing import Collection

from prompt_toolkit import shortcuts
from prompt_toolkit.formatted_text import merge_formatted_text
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.keys import Keys
from prompt_toolkit.validation import Validator

_NO_INPUT = ""


def echo(*args, **kwargs) -> None:
    shortcuts.print_formatted_text(*args, **kwargs)


def acknowledge(*args, no_user_input: bool = False, **kwargs) -> None:
    echo(*args, **kwargs)
    if no_user_input:
        return

    shortcuts.prompt("", **kwargs)
    echo("")


def confirm(
    question: str, default: bool = False, no_user_input: bool = False, **kwargs
) -> bool:
    if no_user_input:
        return default

    answer = _confirm(question, default, **kwargs)
    echo("")

    return answer


def ask(question: str, default: str, no_user_input: bool = False, **kwargs) -> str:
    if no_user_input:
        return default

    # Don't pass real default to prompt as it requires the user to delete the characters to enter something custom
    answer = shortcuts.prompt(f"{question} [{default}]: ", default=_NO_INPUT, **kwargs)
    if answer == _NO_INPUT:
        answer = default
    echo("")

    return answer


def multiple_choice(
    question: str,
    options: Collection[str],
    default: str,
    no_user_input: bool = False,
    **kwargs,
) -> str:
    if len(options) == 0:
        raise ValueError("options must contain at least one value")

    prompt_lines = [question]
    default_choice = None
    choice_map = {}
    for i, value in enumerate(options):
        key = str(i + 1)
        choice_map[key] = value
        prompt_lines.append(f"{key} - {value}")

        if default == value:
            default_choice = key

    if default_choice is None:
        raise ValueError(f"""Default "{default}" was not an option {options}""")

    prompt_lines.append("Enter the number of your choice")

    user_choice = ask(
        "\n".join(prompt_lines),
        validator=Validator.from_callable(
            lambda text: text == _NO_INPUT or text in choice_map.keys()
        ),
        default=default_choice,
        no_user_input=no_user_input,
        **kwargs,
    )

    return choice_map[user_choice]


def _confirm(question: str, default: bool = False, **kwargs) -> bool:
    bindings = KeyBindings()

    @bindings.add("y")
    @bindings.add("Y")
    def _yes(event: KeyPressEvent) -> None:
        session.default_buffer.text = "y"
        event.app.exit(result=True)

    @bindings.add("n")
    @bindings.add("N")
    def _no(event: KeyPressEvent) -> None:
        session.default_buffer.text = "n"
        event.app.exit(result=False)

    @bindings.add(Keys.Any)
    def _any(_: KeyPressEvent) -> None:
        # Disallow inserting other text.
        pass

    if default:
        default_indicator = "Y/n"
        bindings.add(Keys.Enter)(_yes)
    else:
        default_indicator = "y/N"
        bindings.add(Keys.Enter)(_no)

    complete_message = merge_formatted_text([question, f" ({default_indicator}): "])
    session: shortcuts.PromptSession[bool] = shortcuts.PromptSession(
        complete_message, key_bindings=bindings, **kwargs
    )
    return session.prompt(complete_message, key_bindings=bindings, **kwargs)



import pytest

from columbo import (
    Acknowledge,
    BasicQuestion,
    Choice,
    Confirm,
    DuplicateQuestionNameException,
    Echo,
    ValidationFailure,
    ValidationSuccess,
)
from columbo._interaction import canonical_arg_name, get_answers, to_value
from tests.sample_data import (
    DUPLICATE_QUESTION_NAME_PARAMS,
    QUESTION_NAME_STANDARDIZATION_PARAMS,
    SOME_ANSWERS,
    SOME_BOOL,
    SOME_DEFAULT,
    SOME_DYNAMIC_DEFAULT_RESULT,
    SOME_DYNAMIC_OPTION_RESULT,
    SOME_DYNAMIC_STRING_RESULT,
    SOME_NAME,
    SOME_OPTIONS,
    SOME_OTHER_BOOL,
    SOME_STRING,
    SampleQuestion,
    some_dynamic_bool,
    some_dynamic_default,
    some_dynamic_options,
    some_dynamic_string,
)


def test_echo__static_message__echo_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Echo(SOME_STRING).display(SOME_ANSWERS)

    user_io.echo.assert_called_once_with(SOME_STRING)


def test_echo__dynamic_message__echo_dynamic_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Echo(some_dynamic_string).display(SOME_ANSWERS)

    user_io.echo.assert_called_once_with(SOME_DYNAMIC_STRING_RESULT)


def test_echo_copy__new_instance():
    original = Echo(SOME_STRING)

    copy = original.copy()

    assert copy is not original


def test_echo_copy__no_args__echo_same_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    interaction = Echo(SOME_STRING)
    interaction.display(SOME_ANSWERS)

    interaction.copy().display(SOME_ANSWERS)

    calls = user_io.echo.mock_calls
    assert len(calls) == 2
    assert calls[0] == calls[1]


def test_echo_copy__diff_message__echo_diff_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Echo(some_dynamic_string).copy(message=SOME_STRING).display(SOME_ANSWERS)

    user_io.echo.assert_called_once_with(SOME_STRING)


def test_acknowledge__static_message__acknowledge_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Acknowledge(SOME_STRING).display(SOME_ANSWERS)

    user_io.acknowledge.assert_called_once_with(SOME_STRING)


def test_acknowledge__dynamic_message__acknowledge_dynamic_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Acknowledge(some_dynamic_string).display(SOME_ANSWERS)

    user_io.acknowledge.assert_called_once_with(SOME_DYNAMIC_STRING_RESULT)


def test_acknowledge_copy__new_instance():
    original = Acknowledge(SOME_STRING)

    copy = original.copy()

    assert copy is not original


def test_acknowledge_copy__no_args__acknowledge_same_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    interaction = Acknowledge(SOME_STRING)
    interaction.display(SOME_ANSWERS)

    interaction.copy().display(SOME_ANSWERS)

    calls = user_io.acknowledge.mock_calls
    assert len(calls) == 2
    assert calls[0] == calls[1]


def test_acknowledge_copy__diff_message__acknowledge_diff_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Acknowledge(some_dynamic_string).copy(message=SOME_STRING).display(SOME_ANSWERS)

    user_io.acknowledge.assert_called_once_with(SOME_STRING)


def test_choice__no_input__default_value():
    result = Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT).ask(
        SOME_ANSWERS, no_user_input=True
    )

    assert result == SOME_DEFAULT


def test_choice__static_message__multiple_choice_called(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT).ask(
        SOME_ANSWERS, no_user_input=True
    )

    user_io.multiple_choice.assert_called_once_with(
        SOME_STRING, SOME_OPTIONS, default=SOME_DEFAULT, no_user_input=True
    )


def test_choice__dynamic_message__multiple_choice_dynamic_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Choice(
        SOME_NAME, some_dynamic_string, some_dynamic_options, some_dynamic_default
    ).ask(SOME_ANSWERS, no_user_input=True)

    user_io.multiple_choice.assert_called_once_with(
        SOME_DYNAMIC_STRING_RESULT,
        SOME_DYNAMIC_OPTION_RESULT,
        default=SOME_DYNAMIC_DEFAULT_RESULT,
        no_user_input=True,
    )


@pytest.mark.parametrize("value,is_valid", [(SOME_DEFAULT, True), (SOME_STRING, False)])
def test_choice_is_valid__static_options__expected_result(value, is_valid):
    question = Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)

    result = question.validate(value, SOME_ANSWERS)

    assert result.valid == is_valid


@pytest.mark.parametrize(
    "value,is_valid", [(SOME_DYNAMIC_DEFAULT_RESULT, True), (SOME_STRING, False)]
)
def test_choice_is_valid__dynamic_options__expected_result(value, is_valid):
    question = Choice(SOME_NAME, SOME_STRING, some_dynamic_options, SOME_DEFAULT)

    result = question.validate(value, SOME_ANSWERS)

    assert result.valid == is_valid


@pytest.mark.parametrize(
    "value,is_valid", [(SOME_DYNAMIC_DEFAULT_RESULT, True), (SOME_STRING, False)]
)
def test_choice__validate__dynamic_options__expected_result(value, is_valid):
    question = Choice(SOME_NAME, SOME_STRING, some_dynamic_options, SOME_DEFAULT)

    result = question.validate(value, SOME_ANSWERS)

    assert result.valid == is_valid


def test_choice_copy__new_instance():
    original = Choice(
        SOME_NAME, some_dynamic_string, some_dynamic_options, some_dynamic_default
    )

    copy = original.copy()

    assert copy is not original


def test_choice_copy__no_args__multiple_choice_same_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    question = Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT)
    question.ask(SOME_ANSWERS, no_user_input=True)

    question.copy().ask(SOME_ANSWERS, no_user_input=True)

    calls = user_io.multiple_choice.mock_calls
    assert len(calls) == 2
    assert calls[0] == calls[1]


def test_choice_copy__diff_message__multiple_choice_diff_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Choice(
        SOME_NAME, some_dynamic_string, some_dynamic_options, some_dynamic_default
    ).copy(message=SOME_STRING, options=SOME_OPTIONS, default=SOME_DEFAULT).ask(
        SOME_ANSWERS, no_user_input=True
    )

    user_io.multiple_choice.assert_called_once_with(
        SOME_STRING, SOME_OPTIONS, default=SOME_DEFAULT, no_user_input=True
    )


def test_confirm__no_input__default_value():
    result = Confirm(SOME_NAME, SOME_STRING, SOME_BOOL).ask(
        SOME_ANSWERS, no_user_input=True
    )

    assert result == SOME_BOOL


def test_confirm__static_message__confirm_called(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Confirm(SOME_NAME, SOME_STRING, SOME_BOOL).ask(SOME_ANSWERS, no_user_input=True)

    user_io.confirm.assert_called_once_with(
        SOME_STRING, default=SOME_BOOL, no_user_input=True
    )


def test_confirm__dynamic_message__confirm_dynamic_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Confirm(SOME_NAME, some_dynamic_string, some_dynamic_bool).ask(
        SOME_ANSWERS, no_user_input=True
    )

    user_io.confirm.assert_called_once_with(
        SOME_DYNAMIC_STRING_RESULT, default=SOME_OTHER_BOOL, no_user_input=True
    )


def test_confirm_copy__new_instance():
    original = Confirm(SOME_NAME, SOME_STRING, SOME_BOOL)

    copy = original.copy()

    assert copy is not original


def test_confirm_copy__no_args__confirm_same_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    question = Confirm(SOME_NAME, SOME_STRING, SOME_BOOL)
    question.ask(SOME_ANSWERS, no_user_input=True)

    question.copy().ask(SOME_ANSWERS, no_user_input=True)

    calls = user_io.confirm.mock_calls
    assert len(calls) == 2
    assert calls[0] == calls[1]


def test_confirm_copy__diff_message__confirm_diff_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    Confirm(SOME_NAME, some_dynamic_string, some_dynamic_bool).copy(
        message=SOME_STRING, default=SOME_BOOL
    ).ask(SOME_ANSWERS, no_user_input=True)

    user_io.confirm.assert_called_once_with(
        SOME_STRING, default=SOME_BOOL, no_user_input=True
    )


def test_basic_question__no_input__default_value():
    result = BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT).ask(
        SOME_ANSWERS, no_user_input=True
    )

    assert result == SOME_DEFAULT


def test_basic_question__ask__validator__default_value():
    result = BasicQuestion(
        SOME_NAME, SOME_STRING, SOME_DEFAULT, validator=lambda v, a: ValidationSuccess()
    ).ask(SOME_ANSWERS, no_user_input=True)

    assert result == SOME_DEFAULT


def test_basic_question__static_message__ask_called(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT).ask(
        SOME_ANSWERS, no_user_input=True
    )

    user_io.ask.assert_called_once_with(
        SOME_STRING, default=SOME_DEFAULT, no_user_input=True
    )


def test_basic_question__dynamic_message__ask_dynamic_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    BasicQuestion(SOME_NAME, some_dynamic_string, some_dynamic_default).ask(
        SOME_ANSWERS, no_user_input=True
    )

    user_io.ask.assert_called_once_with(
        SOME_DYNAMIC_STRING_RESULT,
        default=SOME_DYNAMIC_DEFAULT_RESULT,
        no_user_input=True,
    )


def test_basic_question_copy__new_instance():
    original = BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)

    copy = original.copy()

    assert copy is not original


def test_basic_question_copy__no_args__confirm_same_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    question = BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)
    question.ask(SOME_ANSWERS, no_user_input=True)

    question.copy().ask(SOME_ANSWERS, no_user_input=True)

    calls = user_io.ask.mock_calls
    assert len(calls) == 2
    assert calls[0] == calls[1]


def test_basic_question_copy__diff_message__confirm_diff_message(mocker):
    user_io = mocker.patch("columbo._interaction.user_io")

    BasicQuestion(SOME_NAME, some_dynamic_string, some_dynamic_default).copy(
        message=SOME_STRING, default=SOME_DEFAULT
    ).ask(SOME_ANSWERS, no_user_input=True)

    user_io.ask.assert_called_once_with(
        SOME_STRING, default=SOME_DEFAULT, no_user_input=True
    )


@pytest.mark.parametrize("value", [SOME_DEFAULT, SOME_STRING])
def test_basic_question_is_valid__no_validator__always_valid(value):
    question = BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)

    result = question.validate(value, SOME_ANSWERS)

    assert result.valid


@pytest.mark.parametrize(
    ["is_valid", "validator_response"],
    [(True, None), (False, "some-error")],
)
def test_basic_question_is_valid__legacy_validator__result_of_validator(
    is_valid, validator_response
):
    question = BasicQuestion(
        SOME_NAME, SOME_STRING, SOME_DEFAULT, validator=lambda v, a: validator_response
    )

    with pytest.deprecated_call():
        result = question.validate(SOME_STRING, SOME_ANSWERS)
        assert result.valid == is_valid


@pytest.mark.parametrize(
    ["is_valid", "validator_response"],
    [(True, ValidationSuccess()), (False, ValidationFailure(error="some-error"))],
)
def test_basic_question_is_valid__validator__result_of_validator(
    is_valid, validator_response
):
    question = BasicQuestion(
        SOME_NAME, SOME_STRING, SOME_DEFAULT, validator=lambda v, a: validator_response
    )

    result = question.validate(SOME_STRING, SOME_ANSWERS)

    assert result.valid == is_valid


def test_basic_question_is_valid__invalid_validator__exception():
    question = BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT, validator=object())

    with pytest.raises(ValueError):
        question.validate(SOME_STRING, SOME_ANSWERS)


@pytest.mark.parametrize(
    "validity_responses",
    [[None], ["some-error", None], ["some-error1", "some-error2", None]],
)
def test_basic_question__invalid_asked_multiple_times(mocker, validity_responses):
    """
    A BasicQuestion will continue to be asked until a valid response is provided

    We force the validator to fail several times, and assert the user_io.ask() is invoked the correct amount of times.
    """

    user_io_ask_mock = mocker.patch("columbo._interaction.user_io.ask")

    with pytest.deprecated_call():
        # validator callable will return False and then True when invoked
        BasicQuestion(
            SOME_NAME,
            SOME_STRING,
            SOME_DEFAULT,
            validator=mocker.Mock(side_effect=validity_responses),
        ).ask(SOME_ANSWERS)

    assert user_io_ask_mock.call_count == len(validity_responses)


def test_to_value__invalid_type__exception():
    with pytest.raises(ValueError):
        to_value(object(), SOME_ANSWERS, str)


@pytest.mark.parametrize(
    "description,should_ask,expected_result",
    [("not set", None, True), ("dynamic", some_dynamic_bool, SOME_OTHER_BOOL)],
)
def test_should_ask__expected_result(should_ask, expected_result, description):
    result = SampleQuestion(SOME_NAME, SOME_STRING, should_ask=should_ask).should_ask(
        SOME_ANSWERS
    )

    assert result == expected_result, description


def test_should_ask__invalid_type__exception():
    with pytest.raises(ValueError):
        SampleQuestion(SOME_NAME, SOME_STRING, should_ask=object()).should_ask(
            SOME_ANSWERS
        )


def test_get_answers__unknown_interaction_provided():
    """
    An unsupported interaction type should raise a ValueError
    """
    # `5` is not an interaction type that is supported
    interactions = [Echo("foo"), 5]
    with pytest.raises(ValueError):
        get_answers(interactions)


@pytest.mark.parametrize(
    ["basic_question_should_ask", "expected_basic_question_ask_call_count"],
    [(True, 1), (False, 0)],
)
def test_get_answers__proper_funcs_called(
    mocker, basic_question_should_ask, expected_basic_question_ask_call_count
):
    """
    Given a list of interactions of different types, validate the proper funcs should be invoked on each

    The parametrization asserts that Question interactions are only 'asked' when they should be asked
    """
    echo_interaction_mock = mocker.Mock(spec=Echo)
    basic_question_interaction_mock = mocker.Mock(
        spec=BasicQuestion,
        should_ask=mocker.Mock(return_value=basic_question_should_ask),
    )
    # mock names must be set after mock object is created
    echo_interaction_mock.name = "echo_interaction_mock"
    basic_question_interaction_mock.name = "basic_question_interaction_mock"

    interactions = [echo_interaction_mock, basic_question_interaction_mock]
    get_answers(interactions)

    # functions belonging to interactions should be invoked
    # in the case of a question, it either should or shouldn't be asked
    echo_interaction_mock.display.assert_called_once()
    basic_question_interaction_mock.should_ask.assert_called_once()
    assert (
        basic_question_interaction_mock.ask.call_count
        == expected_basic_question_ask_call_count
    )


def test_get_answers__previous_answers__result_has_previous_answers_and_question_answers():
    existing_key = "question1"
    existing_answer = "some previous answer"
    existing_answers = {existing_key: existing_answer}
    expected_answers = {existing_key: existing_answer, SOME_NAME: SOME_DEFAULT}

    interactions = [
        BasicQuestion(SOME_NAME, "What is the answer to life?", default=SOME_DEFAULT)
    ]

    new_answers = get_answers(interactions, existing_answers, no_user_input=True)

    assert new_answers == expected_answers


@pytest.mark.parametrize("questions", DUPLICATE_QUESTION_NAME_PARAMS)
def test_parse_args__duplicate_question_name__exception(questions):
    with pytest.raises(DuplicateQuestionNameException):
        get_answers(questions)


def test_parse_args__duplicate_question_name_in_answers__exception():
    with pytest.raises(DuplicateQuestionNameException):
        get_answers(
            [BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT)],
            answers={SOME_NAME: "existing value"},
        )


@pytest.mark.parametrize(
    ["description", "name", "expected_arg"], QUESTION_NAME_STANDARDIZATION_PARAMS
)
def test__canonical_arg_name__nonstandard_characters__normalized(
    description, name, expected_arg
):
    assert canonical_arg_name(name) == expected_arg, description



from argparse import Namespace

from columbo._interaction import (
    BasicQuestion,
    Choice,
    Confirm,
    Question,
    canonical_arg_name,
)

SOME_BOOL = True
SOME_OTHER_BOOL = False
SOME_NAME = "my-test-value"
SOME_NAME_WITH_SPACES = "my test value"
SOME_ARG_NAME = canonical_arg_name(SOME_NAME)
SOME_STRING = "hello"
SOME_OTHER_STRING = "good-bye"
SOME_ANSWERS = {"a": "one", "b": "two"}
SOME_OPTIONS = ["x", "y", "z"]
SOME_DEFAULT = "x"
SOME_NON_DEFAULT_OPTION = "y"
SOME_INVALID_OPTION = "NOT_VALID_OPTION"
SOME_INVALID_ARG_NAME = "--NOT_VALID_OPTION"


def some_dynamic_string(answers):
    return f"--{answers['a']}--"


def some_dynamic_options(answers):
    return [f"--{x}--" for x in answers.values()]


def some_dynamic_default(answers):
    return f"--{answers['b']}--"


def some_dynamic_bool(_):
    return SOME_OTHER_BOOL


SOME_DYNAMIC_STRING_RESULT = "--one--"
SOME_DYNAMIC_OPTION_RESULT = ["--one--", "--two--"]
SOME_DYNAMIC_DEFAULT_RESULT = "--two--"

SOME_NAMESPACE = Namespace(**{SOME_NAME: SOME_STRING})


class SampleQuestion(Question):
    """Question class for testing base class functionality or where subclasses need specific handling"""

    def ask(self, answers, no_user_input=False):
        raise Exception("Don't call")


# combination of questions that reuse the same name
DUPLICATE_QUESTION_NAME_PARAMS = [
    [
        BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT),
        BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT),
    ],
    [
        Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT),
        Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT),
    ],
    [Confirm(SOME_NAME, SOME_STRING), Confirm(SOME_NAME, SOME_STRING)],
    [
        BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT),
        Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT),
    ],
    [
        BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT),
        Confirm(SOME_NAME, SOME_STRING),
    ],
    [
        Choice(SOME_NAME, SOME_STRING, SOME_OPTIONS, SOME_DEFAULT),
        Confirm(SOME_NAME, SOME_STRING),
    ],
    [
        BasicQuestion(SOME_NAME, SOME_STRING, SOME_DEFAULT),
        BasicQuestion(
            SOME_NAME_WITH_SPACES, SOME_OTHER_STRING, SOME_NON_DEFAULT_OPTION
        ),
    ],
]

QUESTION_NAME_STANDARDIZATION_PARAMS = [
    ["basic name", "foo", "--foo"],
    ["name with one leading dash", "-foo", "--foo"],
    ["name with trailing dash", "foo-", "--foo"],
    ["name with trailing and leading spaces", " foo ", "--foo"],
    ["name with trailing and leading underscores", "__foo__", "--foo"],
    ["name with leading dashes", "--foo", "--foo"],
    ["name with spaces", "foo bar", "--foo-bar"],
    ["name with underscore", "foo_bar", "--foo-bar"],
    ["name with double underscore", "foo__bar", "--foo-bar"],
    ["name with double underscore", "foo__bar", "--foo-bar"],
    ["name with mixed capitalization", "Foo-Bar", "--foo-bar"],
    ["name with multiple mixed dashes", "foo--bar__baz", "--foo-bar-baz"],
]



"""
Produce a CLI based on a sequence of interactions.
"""

from argparse import ArgumentParser, Namespace
from functools import singledispatch
from typing import Any, Collection, Dict, NoReturn, Optional, Sequence, Union, cast

from columbo._exception import CliException
from columbo._interaction import (
    Acknowledge,
    BasicQuestion,
    Choice,
    Confirm,
    Echo,
    Interaction,
    canonical_arg_name,
    to_value,
    validate_duplicate_question_names,
)
from columbo._types import Answers, MutableAnswers

CliResult = Union[str, bool]
CliResults = Dict[str, CliResult]


def parse_args(
    interactions: Collection[Interaction],
    args: Optional[Sequence[str]] = None,
    exit_on_error: bool = True,
    answers: Optional[Answers] = None,
    parser_name: Optional[str] = None,
) -> MutableAnswers:
    """
    Parse command line argument for the given interactions.

    :param interactions: Interactions that should be turned into CLI arguments.
    :param args: Arguments to parse. If `None`, `sys.argv` will be used.
    :param exit_on_error: If `True`, print the CLI usage and exit the application. Otherwise, raise an exception with
        the error information.
    :param answers: An initial dictionary of answers to start from.
    :param parser_name: Optional name to be used in error text. If omitted, the name of the process will be used.
    :return: Answers based on the given arguments.
    :raises SystemExit: A value passed to CLI argument was not valid and `exit_on_error` was `True`.
    :raises CliException: A value passed to CLI argument was not valid and `exit_on_error` was `False`.
    :raises DuplicateQuestionNameException: One of the given questions attempts to reuse a name. This includes a
        question that was used to create `answers` if given.
    """
    validate_duplicate_question_names(interactions, answers)
    parser = create_parser(interactions, parser_name)

    if not exit_on_error:
        _patch_parser_error(parser)

    result = parser.parse_args(args)
    try:
        return to_answers(interactions, result, answers)
    except CliException as ex:
        if exit_on_error:
            parser.error(str(ex))
        else:
            raise


def format_cli_help(
    interactions: Collection[Interaction], parser_name: Optional[str] = None
) -> str:
    """
    Produce CLI help text for a given set of interactions.

    :param interactions: Interactions that should be turned into CLI arguments.
    :param parser_name: Optional name to be used in help text. If omitted, the name of the process will be used.
    :raises DuplicateQuestionNameException: One of the given questions attempts to reuse a name.
    """
    validate_duplicate_question_names(interactions)
    return create_parser(interactions, parser_name).format_help()


def create_parser(
    interactions: Collection[Interaction], parser_name: Optional[str] = None
) -> ArgumentParser:
    parser = ArgumentParser(prog=parser_name, add_help=False)
    for interaction in interactions:
        _add_argument_for(interaction, parser)

    return parser


def _patch_parser_error(parser: ArgumentParser) -> None:
    """
    Patch the error method to raise a CliException instead of printing the usage & exiting the application.

    It is safe to just patch error() because the only things that call exit() are the help and version actions, which
    aren't added.

    :param parser: Parser to be patched.
    """

    def _raise_instead(message: str) -> NoReturn:
        raise CliException(message)

    parser.error = _raise_instead  # type: ignore


@singledispatch
def _add_argument_for(question: Any, _: ArgumentParser) -> None:
    raise ValueError(f"Unsupported interaction type {type(question)}")


@_add_argument_for.register(Acknowledge)
@_add_argument_for.register(Echo)
def _add_argument_for_noop(question, parser: ArgumentParser) -> None:
    pass


@_add_argument_for.register(BasicQuestion)
def _add_argument_for_basic(question: BasicQuestion, parser: ArgumentParser) -> None:
    _add_argument(parser, question.name, question.cli_help)


@_add_argument_for.register(Confirm)
def _add_argument_for_confirm(question: Confirm, parser: ArgumentParser) -> None:
    _add_flag(parser, question.name, question.cli_help, active=True)
    _add_flag(parser, question.name, question.cli_help, active=False)


@_add_argument_for.register(Choice)
def _add_argument_for_choice(question: Choice, parser: ArgumentParser) -> None:
    options = question.options
    _add_argument(
        parser,
        question.name,
        question.cli_help,
        # For dynamic options we don't restrict the values in the CLI.
        # Conflicts will be rejected when processing the results.
        choices=None if callable(options) else options,
    )


def to_answers(
    interactions: Collection[Interaction],
    result: Namespace,
    answers: Optional[Answers] = None,
) -> MutableAnswers:
    cli_values: CliResults = vars(result)
    resultant_answers = {} if answers is None else dict(answers)

    for interaction in interactions:
        _update_answers(interaction, cli_values, resultant_answers)

    return resultant_answers


@singledispatch
def _update_answers(
    question: Any, cli_values: CliResults, answers: MutableAnswers
) -> None:
    raise ValueError(f"Unsupported interaction type {type(question)}")


@_update_answers.register(Acknowledge)
@_update_answers.register(Echo)
def _update_answers_noop(
    question, cli_values: CliResults, answers: MutableAnswers
) -> None:
    pass


@_update_answers.register(BasicQuestion)
@_update_answers.register(Choice)
def _update_answers_validate(
    question: Union[BasicQuestion, Choice],
    cli_values: CliResults,
    answers: MutableAnswers,
) -> None:
    if not question.should_ask(answers):
        return
    value = cast(str, cli_values.get(question.name))
    if value is None:
        value = to_value(question.default, answers, str)
    result = question.validate(value, answers)
    if not result.valid:
        raise CliException.invalid_value(
            value, canonical_arg_name(question.name), result.error
        )
    answers[question.name] = value


@_update_answers.register(Confirm)
def _update_answers_confirm(
    question: Confirm, cli_values: CliResults, answers: MutableAnswers
) -> None:
    if not question.should_ask(answers):
        return
    value = cli_values.get(question.name)
    if value is None:
        value = to_value(question.default, answers, bool)
    answers[question.name] = value


def _add_argument(
    parser: ArgumentParser,
    name: str,
    cli_help: Optional[str],
    dest: Optional[str] = None,
    action: str = "store",
    **kwargs,
) -> None:
    parser.add_argument(
        canonical_arg_name(name),
        action=action,
        dest=dest or name,
        help=cli_help,
        **kwargs,
    )


def _add_flag(
    parser: ArgumentParser,
    name: str,
    cli_help: Optional[str],
    active: bool = True,
    **kwargs,
) -> None:
    # store_const is used for boolean values because store_true/store_false automatically set a default for the
    # argument. For consistency across all args, we don't want this.
    _add_argument(
        parser,
        name if active else f"no-{name}",
        cli_help,
        dest=name,
        action="store_const",
        const=active,
        **kwargs,
    )



import pytest

from columbo import _user_io as user_io

SOME_BOOL = True
SOME_OTHER_BOOL = False
SOME_STRING = "hello"
SOME_OTHER_STRING = "good-bye"


def test_acknowledge__yes_user_input__default_value(mocker):
    mock_prompt = mocker.patch("prompt_toolkit.shortcuts.prompt")

    user_io.acknowledge("Some question?")

    mock_prompt.assert_called_once_with("")


def test_acknowledge__no_user_input__default_value(mocker):
    mock_prompt = mocker.patch("prompt_toolkit.shortcuts.prompt")

    user_io.acknowledge("Some question?", no_user_input=True)

    mock_prompt.assert_not_called()


def test_confirm__no_user_input__default_value():
    result = user_io.confirm("Some question?", no_user_input=True, default=SOME_BOOL)

    assert result == SOME_BOOL


@pytest.mark.parametrize("response", [True, False])
def test_confirm__yes_user_input__confirm_result(response, mocker):
    mock_session = mocker.Mock()
    mock_session.prompt.return_value = response
    mocker.patch("prompt_toolkit.shortcuts.PromptSession", return_value=mock_session)

    result = user_io.confirm("Some question?", default=SOME_OTHER_BOOL)

    assert result == response


def test_ask__no_user_input__default_value():
    result = user_io.ask("Some question?", no_user_input=True, default=SOME_STRING)

    assert result == SOME_STRING


def test_ask__yes_user_input__prompt_result(mocker):
    mocker.patch("prompt_toolkit.shortcuts.prompt", return_value=SOME_STRING)

    result = user_io.ask("Some question?", default=SOME_OTHER_STRING)

    assert result == SOME_STRING


def test_ask__yes_user_input_no_answer__default_result(mocker):
    mocker.patch("prompt_toolkit.shortcuts.prompt", return_value="")

    result = user_io.ask("Some question?", default=SOME_OTHER_STRING)

    assert result == SOME_OTHER_STRING


def test_multiple_choice__no_user_input__default_value():
    result = user_io.multiple_choice(
        "Some question?",
        [SOME_STRING, SOME_OTHER_STRING],
        no_user_input=True,
        default=SOME_OTHER_STRING,
    )

    assert result == SOME_OTHER_STRING


def test_multiple_choice__yes_user_input__prompt_result_mapped_to_value(mocker):
    # expected value is second option
    mocker.patch("prompt_toolkit.shortcuts.prompt", return_value="2")

    result = user_io.multiple_choice(
        "Some question?", [SOME_STRING, SOME_OTHER_STRING], default=SOME_STRING
    )

    assert result == SOME_OTHER_STRING


def test_multiple_choice__default_not_option__value_error():
    with pytest.raises(ValueError):
        user_io.multiple_choice("Some question?", ["1", "2", "3"], default="100")


def test_multiple_choice__no_options__value_error():
    with pytest.raises(ValueError):
        user_io.multiple_choice("Some question?", [], default="100")



"""columbo - Specify a dynamic set of questions to ask a user and get their answers."""
from columbo._cli import format_cli_help, parse_args  # noqa: F401
from columbo._exception import (  # noqa: F401
    CliException,
    ColumboException,
    DuplicateQuestionNameException,
)
from columbo._interaction import (  # noqa: F401
    Acknowledge,
    BasicQuestion,
    Choice,
    Confirm,
    Echo,
    Interaction,
    Question,
    get_answers,
)
from columbo._types import (  # noqa: F401
    Answer,
    Answers,
    MutableAnswers,
    OptionList,
    ShouldAsk,
    StaticOrDynamicValue,
    ValidationFailure,
    ValidationResponse,
    ValidationSuccess,
    Validator,
)

__version__ = "0.10.1"
__author__ = "Patrick Lannigan <plannigan@wayfair.com>"
__all__ = []  # type: ignore
